#ifdef _WIN32
#include <windows.h>
#endif

#include "cuda.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <memory.h>

#include <cuda_runtime.h>
#include <cuda_gl_interop.h>

#include "WaterLBComputation.h"

//-------------------------------------------------------------------------------------------

bool getComptableCUDADevices(int **dArrayPtr)
{
	int devCount = 0;
	cudaDeviceProp deviceProp;
	bool res = false;
	
	if(cudaGetDeviceCount(&devCount)==cudaSuccess)
	{
		int i,dCount = 0;
		int *dArray;
		
		dArray = (int *)malloc((devCount+1) * sizeof(int));
		for(i=0;i<devCount;i++)
		{
			if(cudaGetDeviceProperties(&deviceProp,i)==cudaSuccess)
			{
				if(deviceProp.major>=2 || (deviceProp.major==1 && deviceProp.minor>=1))
				{
					dArray[dCount++] = i;
				}
			}
		}
		if(dCount>0)
		{
			*dArrayPtr = dArray;
			dArray[dCount] = -1;
			res = true;
		}
		else
		{
			free(dArray);
		}
	}
	else
	{
		devCount = 0;
	}
	return res;
}

//-------------------------------------------------------------------------------------------

bool isCUDADriverUpToDate(char **eStrPtr)
{
	int driverVersion = 0,runtimeVersion = 0;
	int dVMajor,dVMinor;
	int rVMajor,rVMinor;
	char *eStr;
	bool res = true;

	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);
	
	dVMajor = driverVersion / 1000;
	dVMinor = (driverVersion%100) / 10;
	rVMajor = runtimeVersion / 1000;
	rVMinor = (runtimeVersion%100) / 10;
	if((dVMajor<4 || (dVMajor==4 && dVMinor<2)) || (rVMajor<4 || (rVMajor==4 && rVMinor<2)))
	{
		eStr = (char *)malloc(1024);
		sprintf(eStr,"CUDA Driver Version / Runtime Version: %d.%d / %d.%d",dVMajor,dVMinor,rVMajor,rVMinor);
		*eStrPtr = eStr;
		res = false;
	}
	return res;
}

//-------------------------------------------------------------------------------------------

bool checkCUDAResult(cudaError err)
{
	if(err!=cudaSuccess)
	{
		fprintf(stdout,"CUDA Runtime API error %d: %s.\n",(int)err,cudaGetErrorString(err));
		return false;
	}
	else
	{
		return true;
	}
}

//-------------------------------------------------------------------------------------------

void printDeviceProperties(int devID,cudaDeviceProp *deviceProp)
{
	int driverVersion = 0,runtimeVersion = 0;

	fprintf(stdout,"Device %d: \"%s\"\n",devID,deviceProp->name);
	
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);
	fprintf(stdout,"CUDA Driver Version / Runtime Version: %d.%d / %d.%d\n",driverVersion/1000,(driverVersion%100)/10,runtimeVersion/1000,(runtimeVersion%100)/10);
	fprintf(stdout,"CUDA Capability Major/Minor version number: %d.%d\n",deviceProp->major,deviceProp->minor);
}

//-------------------------------------------------------------------------------------------

bool setupGLDevice()
{
	int *devices;
	bool res = false;
	
	if(getComptableCUDADevices(&devices))
	{
		if(checkCUDAResult(cudaGLSetGLDevice(devices[0])))
		{
			res = true;
		}
	}
	return res;
}

//-------------------------------------------------------------------------------------------

bool isKernelSuccessful()
{
	cudaError err = cudaGetLastError();
	return checkCUDAResult(err);
}

//-------------------------------------------------------------------------------------------

static float __constant__ c_gacl = 9.81f;
static float __constant__ c_omega = 1.0f / 1.5f;

static float __constant__ c_ex0 = 0.0f;
static float __constant__ c_ey0 = 0.0f;

static float __constant__ c_ex1 = 1.0f;
static float __constant__ c_ey1 = 0.0f;

static float __constant__ c_ex2 = 1.0f;
static float __constant__ c_ey2 = 1.0f;

static float __constant__ c_ex3 = 0.0f;
static float __constant__ c_ey3 = 1.0f;

static float __constant__ c_ex4 = -1.0f;
static float __constant__ c_ey4 = 1.0f;

static float __constant__ c_ex5 = -1.0f;
static float __constant__ c_ey5 = 0.0f;

static float __constant__ c_ex6 = -1.0f;
static float __constant__ c_ey6 = -1.0f;

static float __constant__ c_ex7 = 0.0f;
static float __constant__ c_ey7 = -1.0f;

static float __constant__ c_ex8 = 1.0f;
static float __constant__ c_ey8 = -1.0f;

	
static float __constant__ c_minHeight = 0.0495f;
static float __constant__ c_maxHeight = 0.0505f;


//-------------------------------------------------------------------------------------------
// 0, 1, 2,  3,  4,  5,  6,  7,  8,
// 0, 4, 8, 12, 16, 20, 24, 28, 32,
// 0, 36, 72
// 4608 / 4 = 1152 / 128 = 9
// nx * ny is multiple of 1152
//-------------------------------------------------------------------------------------------

/*
void WaterGridModel::resolveHUV(int x,int y)
{
	int i,idx = index(x,y);
	tfloat32 h,u,v;
	
	h = 0.0f;
	u = 0.0f;
	v = 0.0f;
	for(i=0;i<9;i++)
	{
		tfloat32 *f = m_f[i];
		h += f[idx];
		u += m_ex[i] * f[idx];
		v += m_ey[i] * f[idx];
	}
	u = u / h;
	v = v / h;
	
	m_h[idx] = h;
	m_u[idx] = u;
	m_v[idx] = v;
}
*/

//-------------------------------------------------------------------------------------------

__global__ void kResolveHUV(float *fArray, float *hArray, float *uArray, float *vArray)
{
	int idx;
	float h, u, v, f;
	
	idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	
	f = fArray[__mul24(idx, 9) + 0];
	h = f;

	f = fArray[__mul24(idx, 9) + 1];
	h += f;
	u = f;
	
	f = fArray[__mul24(idx, 9) + 2];
	h += f;
	u += f;
	v = f;

	f = fArray[__mul24(idx, 9) + 3];
	h += f;
	v += f;
	
	f = fArray[__mul24(idx, 9) + 4];
	h += f;
	u -= f;
	v += f;
	
	f = fArray[__mul24(idx, 9) + 5];
	h += f;
	u -= f;

	f = fArray[__mul24(idx, 9) + 6];
	h += f;
	u -= f;
	v -= f;

	f = fArray[__mul24(idx, 9) + 7];
	h += f;
	v += -f;

	f = fArray[__mul24(idx, 9) + 8];
	h += f;
	u += f;
	v -= f;
	
	hArray[idx] = h;
	uArray[idx] = u / h;
	vArray[idx] = v / h;
}

//-------------------------------------------------------------------------------------------

bool waterLBResolveHUV(WaterLBComputeData *pData)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	kResolveHUV<<<grid,block>>>(pData->force9DMap, pData->heightMap, pData->uVelocityMap, pData->vVelocityMap);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

/*
void WaterGridModel::computeFEQ(int x,int y)
{
	int i,idx = index(x,y);
	tfloat32 *u = m_u;
	tfloat32 *v = m_v;
	tfloat32 *h = m_h;
	
	for(i=0;i<9;i++)
	{
		tfloat32 *fEQ = m_feq[i];
		
		switch(i)
		{
			case 0:
				fEQ[idx]  = h[idx];
				fEQ[idx] -= ((5.0f * m_gacl * h[idx] * h[idx]) / 6.0f);
				fEQ[idx] -= ((2.0f * h[idx]) / 3.0f) * ((u[idx] * u[idx]) + (v[idx] * v[idx]));
				break;
			
			case 1:
			case 3:
			case 5:
			case 7:
				fEQ[idx]  = (m_gacl * h[idx] * h[idx]) / 6.0f;
				fEQ[idx] += (h[idx] / 3.0f) * ((m_ex[i] * u[idx]) + (m_ey[i] * v[idx]));
				fEQ[idx] += (h[idx] / 2.0f) * ((m_ex[i] * u[idx] * m_ex[i] * u[idx]) + (2.0f * m_ex[i] * u[idx] * m_ey[i] * v[idx]) + (m_ey[i] * v[idx] * m_ey[i] * v[idx]));
				fEQ[idx] -= (h[idx] / 6.0f) * ((u[idx] * u[idx]) + (v[idx] * v[idx]));
				break;
			
			case 2:
			case 4:
			case 6:
			case 8:
				fEQ[idx]  = (m_gacl * h[idx] * h[idx]) / 24.0f;
				fEQ[idx] += (h[idx] / 12.0f) * ((m_ex[i] * u[idx]) + (m_ey[i] * v[idx]));
				fEQ[idx] += (h[idx] / 8.0f) * ((m_ex[i] * u[idx] * m_ex[i] * u[idx]) + (2.0f * m_ex[i] * u[idx] * m_ey[i] * v[idx]) + (m_ey[i] * v[idx] * m_ey[i] * v[idx]));
				fEQ[idx] -= (h[idx] / 24.0f) * ((u[idx] * u[idx]) + (v[idx] * v[idx]));
				break;
		}
	}
}
*/

//-------------------------------------------------------------------------------------------
// Stream 0
//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream0(float u, float v, float h)
{
	float fEQ;	
	fEQ  = h;
	fEQ -= (5.0f * c_gacl * h * h) / 6.0f;
	fEQ -= ((2.0f * h) / 3.0f) * ((u * u) + (v * v));
	return fEQ;
}

//-------------------------------------------------------------------------------------------
// Stream 1, 3, 5, 7
//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQAcross(float u, float v, float h, float ex, float ey)
{
	float fEQ;
	fEQ  = (c_gacl * h * h) / 6.0f;
	fEQ += (h / 3.0f) * ((ex * u) + (ey * v));
	fEQ += (h / 2.0f) * ((ex * u * ex * u) + (2.0f * ex * u * ey * v) + (ey * v * ey * v));
	fEQ -= (h / 6.0f) * ((u * u) + (v * v));
	return fEQ;
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream1(float u, float v, float h)
{
	return kComputeFEQAcross(u, v, h, c_ex1, c_ey1);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream3(float u, float v, float h)
{
	return kComputeFEQAcross(u, v, h, c_ex3, c_ey3);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream5(float u, float v, float h)
{
	return kComputeFEQAcross(u, v, h, c_ex5, c_ey5);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream7(float u, float v, float h)
{
	return kComputeFEQAcross(u, v, h, c_ex7, c_ey7);
}

//-------------------------------------------------------------------------------------------
// Stream 2, 4, 6, 8
//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQDiagonal(float u, float v, float h, float ex, float ey)
{
	float fEQ;
	fEQ  = (c_gacl * h * h) / 24.0f;
	fEQ += (h / 12.0f) * ((ex * u) + (ey * v));
	fEQ += (h / 8.0f) * ((ex * u * ex * u) + (2.0f * ex * u * ey * v) + (ey * v * ey * v));
	fEQ -= (h / 24.0f) * ((u * u) + (v * v));
	return fEQ;
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream2(float u, float v, float h)
{
	return kComputeFEQDiagonal(u, v, h, c_ex2, c_ey2);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream4(float u, float v, float h)
{
	return kComputeFEQDiagonal(u, v, h, c_ex4, c_ey4);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream6(float u, float v, float h)
{
	return kComputeFEQDiagonal(u, v, h, c_ex6, c_ey6);
}

//-------------------------------------------------------------------------------------------

__device__ float kComputeFEQStream8(float u, float v, float h)
{
	return kComputeFEQDiagonal(u, v, h, c_ex8, c_ey8);
}

//-------------------------------------------------------------------------------------------

__global__ void kComputeFEQ(float *hArray, float *uArray, float *vArray,float *fEQArray)
{
	int idx;
	float u, v, h;
	
	idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	
	h = hArray[idx];
	u = uArray[idx];
	v = vArray[idx];
	
	fEQArray[__mul24(idx, 9) + 0] = kComputeFEQStream0(u, v, h);
	fEQArray[__mul24(idx, 9) + 1] = kComputeFEQStream1(u, v, h);
	fEQArray[__mul24(idx, 9) + 2] = kComputeFEQStream2(u, v, h);
	fEQArray[__mul24(idx, 9) + 3] = kComputeFEQStream3(u, v, h);
	fEQArray[__mul24(idx, 9) + 4] = kComputeFEQStream4(u, v, h);
	fEQArray[__mul24(idx, 9) + 5] = kComputeFEQStream5(u, v, h);
	fEQArray[__mul24(idx, 9) + 6] = kComputeFEQStream6(u, v, h);
	fEQArray[__mul24(idx, 9) + 7] = kComputeFEQStream7(u, v, h);
	fEQArray[__mul24(idx, 9) + 8] = kComputeFEQStream8(u, v, h);
}

//-------------------------------------------------------------------------------------------

bool waterLBComputeFEQ(WaterLBComputeData *pData)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	kComputeFEQ<<<grid,block>>>(pData->heightMap, pData->uVelocityMap, pData->vVelocityMap, pData->fEQ9DMap);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

/*
void WaterGridModel::collision(bool resolveFlag,int x,int y)
{
	int i,idx = index(x,y);
	
	if(!resolveFlag)
	{
		for(i=0;i<9;i++)
		{
			tfloat32 *f = m_f[i];
			tfloat32 *fEQ = m_feq[i];
			f[idx] = fEQ[idx];
		}
	}

	for(i=0;i<9;i++)
	{
		tfloat32 *f = m_f[i];
		tfloat32 *fEQ = m_feq[i];
		tfloat32 w = (i==0) ? 0.0f : (1.0f / 6.0f);
		f[idx] = ((1.0f - m_omega) * f[idx]) + (m_omega * fEQ[idx]) + (w * ((m_ex[i] * m_forceX[idx]) + (m_ey[i] * m_forceY[idx])));
	}
}
*/

//-------------------------------------------------------------------------------------------

__device__ float kCollisionStream(float f, float fEQ, float forceX, float forceY, float w, float ex, float ey)
{
	return ((1.0f - c_omega) * f) + (c_omega * fEQ) + (w * ((ex * forceX) + (ey * forceY)));
}

//-------------------------------------------------------------------------------------------

__global__ void kCollision(float *fArray, float *fEQArray, float *forceX, float *forceY)
{
	int idx, mIdx;
	float fX, fY;
	
	idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	fX = forceX[idx];
	fY = forceY[idx];
	
	mIdx = __mul24(idx, 9) + 0;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 0.0f, c_ex0, c_ey0);
	mIdx = __mul24(idx, 9) + 1;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex1, c_ey1);
	mIdx = __mul24(idx, 9) + 2;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex2, c_ey2);
	mIdx = __mul24(idx, 9) + 3;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex3, c_ey3);
	mIdx = __mul24(idx, 9) + 4;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex4, c_ey4);
	mIdx = __mul24(idx, 9) + 5;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex5, c_ey5);
	mIdx = __mul24(idx, 9) + 6;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex6, c_ey6);
	mIdx = __mul24(idx, 9) + 7;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex7, c_ey7);
	mIdx = __mul24(idx, 9) + 8;
	fArray[mIdx] = kCollisionStream(fArray[mIdx], fEQArray[mIdx], fX, fY, 1.0f / 6.0f, c_ex8, c_ey8);
}

//-------------------------------------------------------------------------------------------

bool waterLBCollision(WaterLBComputeData *pData, bool resolveFlag)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	
	if(!resolveFlag)
	{
		cudaMemcpy(pData->force9DMap, pData->fEQ9DMap, pData->nx * pData->ny * 9 * sizeof(float), cudaMemcpyDeviceToDevice);
	}
	kCollision<<<grid,block>>>(pData->force9DMap, pData->fEQ9DMap, pData->forceX, pData->forceY);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

bool waterLBCollisionProcess(WaterLBComputeData *pData, bool resolveFlag)
{
	if(resolveFlag)
	{
		if(!waterLBResolveHUV(pData))
		{
			return false;
		}
	}

	if(waterLBComputeFEQ(pData))
	{
		if(waterLBCollision(pData, resolveFlag))
		{
			return true;
		}
	}
	return false;
}

//-------------------------------------------------------------------------------------------

__global__ void kStream(float *fArray, float *fIncArray)
{
	__shared__ float fMem[18][18][9];

	int i, idx;

	int nx = gridDim.x << 4;
	int ny = gridDim.y << 4;
	
	int ix = (threadIdx.x - 1) + (blockIdx.x << 4);
	int iy = (threadIdx.y - 1) + (blockIdx.y << 4);
	
	if(ix>=0 && ix<nx && iy>=0 && iy<ny)
	{
		idx = __mul24(ix + __mul24(nx, iy), 9);
	
		for(i = 0; i < 9; i++)
		{
			fMem[threadIdx.y][threadIdx.x][i] = fArray[idx + i];
		}
	}
	__syncthreads();
	
	if(threadIdx.x < 16 && threadIdx.y < 16)
	{
		ix++;
		iy++;
		
		idx = __mul24(ix + __mul24(nx, iy), 9);
		
		int jx = threadIdx.x + 1;
		int jy = threadIdx.y + 1;
		
		// Stream 0
		fIncArray[idx + 0] = fMem[jy][jx][0];
		
		// Stream 1
		if(ix == 0)
		{
			fIncArray[idx + 1] = fMem[jy][jx][5];
		}
		else
		{
			fIncArray[idx + 1] = fMem[jy][jx-1][1];
		}
		
		// Stream 2
		if(iy == 0 || ix == 0)
		{
			fIncArray[idx + 2] = fMem[jy][jx][6];
		}
		else
		{
			fIncArray[idx + 2] = fMem[jy-1][jx-1][2];
		}
		
		// Stream 3
		if(iy == 0)
		{
			fIncArray[idx + 3] = fMem[jy][jx][7];
		}
		else
		{
			fIncArray[idx + 3] = fMem[jy-1][jx][3];
		}
		
		// Stream 4
		if(iy == 0 || ix == (nx - 1))
		{
			fIncArray[idx + 4] = fMem[jy][jx][8];
		}
		else
		{
			fIncArray[idx + 4] = fMem[jy-1][jx+1][4];
		}
		
		// Stream 5
		if(ix == (nx - 1))
		{
			fIncArray[idx + 5] = fMem[jy][jx][1];
		}
		else
		{
			fIncArray[idx + 5] = fMem[jy][jx+1][5];
		}
		
		// Stream 6
		if(iy == (ny - 1) || ix == (nx - 1))
		{
			fIncArray[idx + 6] = fMem[jy][jx][2];
		}
		else
		{
			fIncArray[idx + 6] = fMem[jy+1][jx+1][6];
		}
		
		// Stream 7
		if(iy == (ny - 1))
		{
			fIncArray[idx + 7] = fMem[jy][jx][3];
		}
		else
		{
			fIncArray[idx + 7] = fMem[jy+1][jx][7];
		}
		
		// Stream 8
		if(iy == (ny - 1) || ix == 0)
		{
			fIncArray[idx + 8] = fMem[jy][jx][4];
		}
		else
		{
			fIncArray[idx + 8] = fMem[jy+1][jx-1][8];
		}
	}
}

//-------------------------------------------------------------------------------------------

bool waterLBStream(WaterLBComputeData *pData)
{
	dim3 block(18,18);
	dim3 grid(pData->nx / 16, pData->ny / 16);
	kStream<<<grid,block>>>(pData->force9DMap, pData->forceInc9DMap);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

void waterLBSwap(WaterLBComputeData *pData)
{
	float *t = pData->force9DMap;
	pData->force9DMap = pData->forceInc9DMap;
	pData->forceInc9DMap = t;
}

//-------------------------------------------------------------------------------------------

/*
void WaterGridModel::incrementForce(tfloat32 *force)
{
	int i,mSize;
	
	mSize = m_nx * m_ny;
	for(i=0;i<mSize;i++)
	{
		const tfloat32 dForce = 0.00001f;

		if(force[i] > dForce)
		{
			force[i] -= dForce;
		}
		else if(force[i] < -dForce)
		{
			force[i] += dForce;
		}
		else
		{
			force[i] = 0.0f;
		}
	}
}
*/

//-------------------------------------------------------------------------------------------

__global__ void kIncrementForce(float *force)
{
	const float dForce = 0.00001f;
	
	int idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	float f = force[idx];
	
	if(f > dForce)
	{
		f -= dForce;
	}
	else if(f < -dForce)
	{
		f += dForce;
	}
	else
	{
		f = 0.0f;
	}
	force[idx] = f;
}

//-------------------------------------------------------------------------------------------

bool waterLBIncrementForce(WaterLBComputeData *pData)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	kIncrementForce<<<grid,block>>>(pData->forceX);
	if(!isKernelSuccessful())
	{
		return false;
	}
	kIncrementForce<<<grid,block>>>(pData->forceY);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

/*
void WaterGridModel::drawPoint(int x,int y,tfloat32 *image,tfloat32 v)
{
	if(x>=0 && x<m_nx && y>=0 && y<m_ny)
	{
		image[x + (y * m_nx)] = v;
	}
}

void WaterGridModel::drawCircleForce(int x0,int y0,tfloat32 radius,tfloat32 v)
{
	int ix,iy,rTop;

	rTop = static_cast<int>(::ceil(radius));
	for(iy=y0-rTop;iy<=y0+rTop;iy++)
	{
		for(ix=x0-rTop;ix<=x0+rTop;ix++)
		{
			tfloat32 dx = static_cast<tfloat32>(ix - x0);
			tfloat32 dy = static_cast<tfloat32>(iy - y0);
			tfloat32 dist = static_cast<tfloat32>(sqrt((dx * dx) + (dy * dy)));
			if(dist <= radius)
			{
				tfloat32 fV = v * ((radius - dist) / radius);
				tfloat32 forceX = fV * (dx / radius);
				tfloat32 forceY = fV * (dy / radius);
				drawPoint(ix,iy,m_forceX,forceX);
				drawPoint(ix,iy,m_forceY,forceY);
			}
		}
	}
}

*/

__global__ void kRainDrop(int x0, int y0, float radius, float v, float *forceX, float *forceY, int lenx, int leny)
{
	// size of tile in which circle force is applied
	int nx = __mul24(blockDim.x, gridDim.x);
	int ny = __mul24(blockDim.y, gridDim.y);

	// coordinate in the tile for given thread
	int ix = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	int iy = threadIdx.y + __mul24(blockIdx.y, blockDim.y);
	
	// (x0,y0) is the centre point of the tile
	// (cx,cy) is the coordinate on the grid for the given thread
	int cx = (ix - (nx >> 1)) + x0;
	int cy = (iy - (ny >> 1)) + y0;
	
	if(cx >= 0 && cx < lenx && cy >= 0 && cy < leny)
	{
		float dx = (float)(cx - x0);
		float dy = (float)(cy - y0);
		float dist = sqrtf((dx * dx) + (dy * dy));
		if(dist <= radius)
		{
			int idx = cx + __mul24(cy, lenx);
			float fV = v * ((radius - dist) / radius);
			
			float fX = fV * (dx / radius);
			forceX[idx] = fX;
			
			float fY = fV * (dy / radius);
			forceY[idx] = fY;
		}
	}
}

//-------------------------------------------------------------------------------------------

bool waterLBRaindropCircleForce(WaterLBComputeData *pData, float posX, float posY, float rForce, float rRadius)
{
	int x0 = (int)(roundf(posX));
	int y0 = (int)(roundf(posY));
	int gridLength = (int)(floorf((rRadius * 2.0f) / 16.0f)) + 1;
	
	dim3 block(16,16);
	dim3 grid(gridLength,gridLength);
	kRainDrop<<<grid,block>>>(x0, y0, rRadius, rForce, pData->forceX, pData->forceY, pData->nx, pData->ny);
	
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

__global__ void kBreakpoint(float *x)
{
	int idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	x[idx] = 0.0f;
}

void waterLBBreakpoint(WaterLBComputeData *pData)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	kBreakpoint<<<grid,block>>>(pData->heightMap);
}

/*
void WaterGridModel::initGrid(int nx,int ny,tfloat32 depth)
{
	int i,mSize = nx * ny;
	
	m_nx = nx;
	m_ny = ny;
	m_depth = depth;
	
	m_step = 0;
	
	freeGrid();
	
	for(i=0;i<9;i++)
	{
		m_feq[i] = new tfloat32 [mSize];
		m_f[i] = new tfloat32 [mSize];
		m_fInc[i] = new tfloat32 [mSize];
	}
	
	m_h = new tfloat32 [mSize];
	m_u = new tfloat32 [mSize];
	m_v = new tfloat32 [mSize];
	m_forceX = new tfloat32 [mSize];
	m_forceY = new tfloat32 [mSize];
	
	m_outputHeightMap = new tuint8 [m_nx * m_ny * 4];
	
	for(i=0;i<mSize;i++)
	{
		m_forceX[i] = 0.0f;
		m_forceY[i] = 0.0f;
		m_h[i] = 0.05f;
		m_u[i] = 0.0f;
		m_v[i] = 0.0f;
	}
}
*/

//-------------------------------------------------------------------------------------------

__global__ void kInitWaterLB(float *hMap, float *uMap, float *vMap, float *forceXMap, float *forceYMap)
{
	int idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	hMap[idx] = 0.05f;
	uMap[idx] = 0.0f;
	vMap[idx] = 0.0f;
	forceXMap[idx] = 0.0f;
	forceYMap[idx] = 0.0f;
}

//-------------------------------------------------------------------------------------------

WaterLBComputeData *waterLBAllocateAndInit(int nx, int ny)
{
	int size = nx * ny;
	cudaError err;
	WaterLBComputeData *pData = 0;
	bool res = false;
	
	// size multiple of 1152
	// nx and ny multiple of 16
	// min size used for test nx = 48, ny = 48 => s = 48*48 = 2304
	
	if((nx & 0xf) || (ny & 0xf) || (size % 1152))
	{
		return 0;
	}
			
	pData = reinterpret_cast<WaterLBComputeData *>(malloc(sizeof(WaterLBComputeData)));
	if(pData != 0)
	{	
		memset(pData, 0, sizeof(WaterLBComputeData));
		
		pData->nx = nx;
		pData->ny = ny;
		
		err = cudaMalloc((void **)&pData->heightMap, size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->uVelocityMap, size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->vVelocityMap, size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->forceX, size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->forceY, size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;

		err = cudaMalloc((void **)&pData->force9DMap, 9 * size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->forceInc9DMap, 9 * size * sizeof(float));
		if(!checkCUDAResult(err))
			goto bail;
		err = cudaMalloc((void **)&pData->fEQ9DMap, 9 * size * sizeof(float));
		
		{	
			dim3 block(576);
			dim3 grid((nx * ny) / 576);
			kInitWaterLB<<<grid, block>>>(pData->heightMap, pData->uVelocityMap, pData->vVelocityMap, pData->forceX, pData->forceY);
			if(!isKernelSuccessful())
				goto bail;
		}
		
		res = true;
	}

bail:
	if(!res)
	{
		waterLBFree(pData);
		pData = 0;
	}
	return pData;
}

//-------------------------------------------------------------------------------------------

void waterLBFree(WaterLBComputeData *pData)
{
	if(pData != 0)
	{
		if(pData->heightMap != 0)
			checkCUDAResult(cudaFree((void *)pData->heightMap));
		if(pData->uVelocityMap != 0)
			checkCUDAResult(cudaFree((void *)pData->uVelocityMap));
		if(pData->vVelocityMap != 0)
			checkCUDAResult(cudaFree((void *)pData->vVelocityMap));
		if(pData->forceX != 0)
			checkCUDAResult(cudaFree((void *)pData->forceX));
		if(pData->forceY != 0)
			checkCUDAResult(cudaFree((void *)pData->forceY));

		if(pData->force9DMap != 0)
			checkCUDAResult(cudaFree((void *)pData->force9DMap));
		if(pData->forceInc9DMap != 0)
			checkCUDAResult(cudaFree((void *)pData->forceInc9DMap));
		if(pData->fEQ9DMap != 0)
			checkCUDAResult(cudaFree((void *)pData->fEQ9DMap));
		
		if(pData->imageData != 0)
			checkCUDAResult(cudaFree((void *)pData->imageData));
		
		if(pData->pboId != 0)
			checkCUDAResult(cudaGraphicsUnregisterResource(pData->glResource));
		
		free(pData);
	}
}

//-------------------------------------------------------------------------------------------

void waterLBGetHeightMap(WaterLBComputeData *pData, float *h)
{
	//waterLBResolveHUV(pData);
	checkCUDAResult(cudaMemcpy(h, pData->heightMap, pData->nx * pData->ny * sizeof(float), cudaMemcpyDeviceToHost));
}

//-------------------------------------------------------------------------------------------

void waterLBInterleave9D(float **src,float *dest,int size)
{
	int i,j;
	
	for(i=0;i<size;i++)
	{
		for(j=0;j<9;j++)
		{
			float *t = src[j];
			dest[(i * 9) + j] = t[i]; 
		}
	}	
}

//-------------------------------------------------------------------------------------------

void waterLBPopulateModel(WaterLBComputeData *pData, float **fEQ, float **f, float **fInc, float *h, float *u, float *v, float *forceX, float *forceY)
{
	int size = pData->nx * pData->ny;
	float *tD9;
	
	tD9 = (float *)malloc(size * 9 * sizeof(float));
	if(tD9)
	{
		waterLBInterleave9D(fEQ, tD9, size);
		checkCUDAResult(cudaMemcpy(pData->fEQ9DMap, tD9, size * 9 * sizeof(float), cudaMemcpyHostToDevice));

		waterLBInterleave9D(f, tD9, size);
		checkCUDAResult(cudaMemcpy(pData->force9DMap, tD9, size * 9 * sizeof(float), cudaMemcpyHostToDevice));

		waterLBInterleave9D(fInc, tD9, size);
		checkCUDAResult(cudaMemcpy(pData->forceInc9DMap, tD9, size * 9 * sizeof(float), cudaMemcpyHostToDevice));
		
		checkCUDAResult(cudaMemcpy(pData->heightMap, h, size * sizeof(float), cudaMemcpyHostToDevice));
		
		checkCUDAResult(cudaMemcpy(pData->uVelocityMap, u, size * sizeof(float), cudaMemcpyHostToDevice));
		
		checkCUDAResult(cudaMemcpy(pData->vVelocityMap, v, size * sizeof(float), cudaMemcpyHostToDevice));
		
		checkCUDAResult(cudaMemcpy(pData->forceX, forceX, size * sizeof(float), cudaMemcpyHostToDevice));
		
		checkCUDAResult(cudaMemcpy(pData->forceY, forceY, size * sizeof(float), cudaMemcpyHostToDevice));
		
		free(tD9);
	}
}

//-------------------------------------------------------------------------------------------

__global__ void kCollision2(float *fArray, float *forceX, float *forceY, float *height, bool resolveFlag)
{
	__shared__ float fMem[8][8][9][2];
	
	float u, v, h;
	
	int ix = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	int iy = threadIdx.y + __mul24(blockIdx.y, blockDim.y);
	int idx = ix + __mul24(iy, __mul24(blockDim.x,gridDim.x));
	int idxF = __mul24(idx, 9);
	
	if(resolveFlag)
	{
		fMem[threadIdx.x][threadIdx.y][0][0] = fArray[idxF + 0];
		h = fMem[threadIdx.x][threadIdx.y][0][0];

		fMem[threadIdx.x][threadIdx.y][1][0] = fArray[idxF + 1];
		h += fMem[threadIdx.x][threadIdx.y][1][0];
		u =  fMem[threadIdx.x][threadIdx.y][1][0];
	
		fMem[threadIdx.x][threadIdx.y][2][0] = fArray[idxF + 2];
		h += fMem[threadIdx.x][threadIdx.y][2][0];
		u += fMem[threadIdx.x][threadIdx.y][2][0];
		v =  fMem[threadIdx.x][threadIdx.y][2][0];

		fMem[threadIdx.x][threadIdx.y][3][0] = fArray[idxF + 3];
		h += fMem[threadIdx.x][threadIdx.y][3][0];
		v += fMem[threadIdx.x][threadIdx.y][3][0];
	
		fMem[threadIdx.x][threadIdx.y][4][0] = fArray[idxF + 4];
		h += fMem[threadIdx.x][threadIdx.y][4][0];
		u -= fMem[threadIdx.x][threadIdx.y][4][0];
		v += fMem[threadIdx.x][threadIdx.y][4][0];
	
		fMem[threadIdx.x][threadIdx.y][5][0] = fArray[idxF + 5];
		h += fMem[threadIdx.x][threadIdx.y][5][0];
		u -= fMem[threadIdx.x][threadIdx.y][5][0];

		fMem[threadIdx.x][threadIdx.y][6][0] = fArray[idxF + 6];
		h += fMem[threadIdx.x][threadIdx.y][6][0];
		u -= fMem[threadIdx.x][threadIdx.y][6][0];
		v -= fMem[threadIdx.x][threadIdx.y][6][0];

		fMem[threadIdx.x][threadIdx.y][7][0] = fArray[idxF + 7];
		h += fMem[threadIdx.x][threadIdx.y][7][0];
		v -= fMem[threadIdx.x][threadIdx.y][7][0];

		fMem[threadIdx.x][threadIdx.y][8][0] = fArray[idxF + 8];
		h += fMem[threadIdx.x][threadIdx.y][8][0];
		u += fMem[threadIdx.x][threadIdx.y][8][0];
		v -= fMem[threadIdx.x][threadIdx.y][8][0];
	
		u = u / h;
		v = v / h;		
	}
	else
	{
		u = 0.0f;
		v = 0.0f;
		h = 0.05f;
	}
	height[idx] = h;
	
	fMem[threadIdx.x][threadIdx.y][0][1] = kComputeFEQStream0(u, v, h);
	fMem[threadIdx.x][threadIdx.y][1][1] = kComputeFEQStream1(u, v, h);
	fMem[threadIdx.x][threadIdx.y][2][1] = kComputeFEQStream2(u, v, h);
	fMem[threadIdx.x][threadIdx.y][3][1] = kComputeFEQStream3(u, v, h);
	fMem[threadIdx.x][threadIdx.y][4][1] = kComputeFEQStream4(u, v, h);
	fMem[threadIdx.x][threadIdx.y][5][1] = kComputeFEQStream5(u, v, h);
	fMem[threadIdx.x][threadIdx.y][6][1] = kComputeFEQStream6(u, v, h);
	fMem[threadIdx.x][threadIdx.y][7][1] = kComputeFEQStream7(u, v, h);
	fMem[threadIdx.x][threadIdx.y][8][1] = kComputeFEQStream8(u, v, h);
	
	if(!resolveFlag)
	{
		fMem[threadIdx.x][threadIdx.y][0][0] = fMem[threadIdx.x][threadIdx.y][0][1];
		fMem[threadIdx.x][threadIdx.y][1][0] = fMem[threadIdx.x][threadIdx.y][1][1];
		fMem[threadIdx.x][threadIdx.y][2][0] = fMem[threadIdx.x][threadIdx.y][2][1];
		fMem[threadIdx.x][threadIdx.y][3][0] = fMem[threadIdx.x][threadIdx.y][3][1];
		fMem[threadIdx.x][threadIdx.y][4][0] = fMem[threadIdx.x][threadIdx.y][4][1];
		fMem[threadIdx.x][threadIdx.y][5][0] = fMem[threadIdx.x][threadIdx.y][5][1];
		fMem[threadIdx.x][threadIdx.y][6][0] = fMem[threadIdx.x][threadIdx.y][6][1];
		fMem[threadIdx.x][threadIdx.y][7][0] = fMem[threadIdx.x][threadIdx.y][7][1];
		fMem[threadIdx.x][threadIdx.y][8][0] = fMem[threadIdx.x][threadIdx.y][8][1];
	}
	
	float fX = forceX[idx];
	float fY = forceY[idx];
	
	fArray[idxF + 0] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][0][0], fMem[threadIdx.x][threadIdx.y][0][1], fX, fY, 0.0f, c_ex0, c_ey0);
	fArray[idxF + 1] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][1][0], fMem[threadIdx.x][threadIdx.y][1][1], fX, fY, 1.0f / 6.0f, c_ex1, c_ey1);
	fArray[idxF + 2] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][2][0], fMem[threadIdx.x][threadIdx.y][2][1], fX, fY, 1.0f / 6.0f, c_ex2, c_ey2);
	fArray[idxF + 3] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][3][0], fMem[threadIdx.x][threadIdx.y][3][1], fX, fY, 1.0f / 6.0f, c_ex3, c_ey3);
	fArray[idxF + 4] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][4][0], fMem[threadIdx.x][threadIdx.y][4][1], fX, fY, 1.0f / 6.0f, c_ex4, c_ey4);
	fArray[idxF + 5] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][5][0], fMem[threadIdx.x][threadIdx.y][5][1], fX, fY, 1.0f / 6.0f, c_ex5, c_ey5);
	fArray[idxF + 6] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][6][0], fMem[threadIdx.x][threadIdx.y][6][1], fX, fY, 1.0f / 6.0f, c_ex6, c_ey6);
	fArray[idxF + 7] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][7][0], fMem[threadIdx.x][threadIdx.y][7][1], fX, fY, 1.0f / 6.0f, c_ex7, c_ey7);
	fArray[idxF + 8] = kCollisionStream(fMem[threadIdx.x][threadIdx.y][8][0], fMem[threadIdx.x][threadIdx.y][8][1], fX, fY, 1.0f / 6.0f, c_ex8, c_ey8);
}

//-------------------------------------------------------------------------------------------

bool waterLBCollisionProcess2(WaterLBComputeData *pData, bool resolveFlag)
{
	dim3 block(8, 8);
	dim3 grid(pData->nx / 8, pData->ny / 8);
	kCollision2<<<grid,block>>>(pData->force9DMap, pData->forceX, pData->forceY, pData->heightMap, resolveFlag);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

__device__ float kRelativeHeight(float h, float min, float max)
{
	float v;
	
	if(h >= min)
	{
		if(h <= max)
		{
			v = (h - min) / (max - min);
		}
		else
		{
			v = 1.0f;
		}
	}
	else
	{
		v = 0.0f;
	}
	return v;
}

__global__ void kCollision3(float *fArray, float *forceX, float *forceY, float *height, bool resolveFlag)
{
	float fMem[9][2];
	
	float u, v, h;
	
	int idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	int idxF = __mul24(idx, 9);
	
	if(resolveFlag)
	{
		fMem[0][0] = fArray[idxF + 0];
		h = fMem[0][0];

		fMem[1][0] = fArray[idxF + 1];
		h += fMem[1][0];
		u =  fMem[1][0];
	
		fMem[2][0] = fArray[idxF + 2];
		h += fMem[2][0];
		u += fMem[2][0];
		v =  fMem[2][0];

		fMem[3][0] = fArray[idxF + 3];
		h += fMem[3][0];
		v += fMem[3][0];
	
		fMem[4][0] = fArray[idxF + 4];
		h += fMem[4][0];
		u -= fMem[4][0];
		v += fMem[4][0];
	
		fMem[5][0] = fArray[idxF + 5];
		h += fMem[5][0];
		u -= fMem[5][0];

		fMem[6][0] = fArray[idxF + 6];
		h += fMem[6][0];
		u -= fMem[6][0];
		v -= fMem[6][0];

		fMem[7][0] = fArray[idxF + 7];
		h += fMem[7][0];
		v -= fMem[7][0];

		fMem[8][0] = fArray[idxF + 8];
		h += fMem[8][0];
		u += fMem[8][0];
		v -= fMem[8][0];
	
		u = u / h;
		v = v / h;		
	}
	else
	{
		u = 0.0f;
		v = 0.0f;
		h = 0.05f;
	}
	
	height[idx] = h;
	
	fMem[0][1] = kComputeFEQStream0(u, v, h);
	fMem[1][1] = kComputeFEQStream1(u, v, h);
	fMem[2][1] = kComputeFEQStream2(u, v, h);
	fMem[3][1] = kComputeFEQStream3(u, v, h);
	fMem[4][1] = kComputeFEQStream4(u, v, h);
	fMem[5][1] = kComputeFEQStream5(u, v, h);
	fMem[6][1] = kComputeFEQStream6(u, v, h);
	fMem[7][1] = kComputeFEQStream7(u, v, h);
	fMem[8][1] = kComputeFEQStream8(u, v, h);
	
	if(!resolveFlag)
	{
		fMem[0][0] = fMem[0][1];
		fMem[1][0] = fMem[1][1];
		fMem[2][0] = fMem[2][1];
		fMem[3][0] = fMem[3][1];
		fMem[4][0] = fMem[4][1];
		fMem[5][0] = fMem[5][1];
		fMem[6][0] = fMem[6][1];
		fMem[7][0] = fMem[7][1];
		fMem[8][0] = fMem[8][1];
	}
	
	float fX = forceX[idx];
	float fY = forceY[idx];
	
	fArray[idxF + 0] = kCollisionStream(fMem[0][0], fMem[0][1], fX, fY, 0.0f, c_ex0, c_ey0);
	fArray[idxF + 1] = kCollisionStream(fMem[1][0], fMem[1][1], fX, fY, 1.0f / 6.0f, c_ex1, c_ey1);
	fArray[idxF + 2] = kCollisionStream(fMem[2][0], fMem[2][1], fX, fY, 1.0f / 6.0f, c_ex2, c_ey2);
	fArray[idxF + 3] = kCollisionStream(fMem[3][0], fMem[3][1], fX, fY, 1.0f / 6.0f, c_ex3, c_ey3);
	fArray[idxF + 4] = kCollisionStream(fMem[4][0], fMem[4][1], fX, fY, 1.0f / 6.0f, c_ex4, c_ey4);
	fArray[idxF + 5] = kCollisionStream(fMem[5][0], fMem[5][1], fX, fY, 1.0f / 6.0f, c_ex5, c_ey5);
	fArray[idxF + 6] = kCollisionStream(fMem[6][0], fMem[6][1], fX, fY, 1.0f / 6.0f, c_ex6, c_ey6);
	fArray[idxF + 7] = kCollisionStream(fMem[7][0], fMem[7][1], fX, fY, 1.0f / 6.0f, c_ex7, c_ey7);
	fArray[idxF + 8] = kCollisionStream(fMem[8][0], fMem[8][1], fX, fY, 1.0f / 6.0f, c_ex8, c_ey8);
}

//-------------------------------------------------------------------------------------------

bool waterLBCollisionProcess3(WaterLBComputeData *pData, bool resolveFlag)
{
	int s = pData->nx * pData->ny;
	dim3 block(576);
	dim3 grid(s / 576);
	kCollision3<<<grid,block>>>(pData->force9DMap, pData->forceX, pData->forceY, pData->heightMap, resolveFlag);
	return isKernelSuccessful();
}

//-------------------------------------------------------------------------------------------

__device__ float kClamp(float x, float a, float b)
{
	return max(a, min(b, x));
}

//-------------------------------------------------------------------------------------------

__device__ int kClamp(int x, int a, int b)
{
	return max(a, min(b, x));
}

//-------------------------------------------------------------------------------------------

__device__ int kRGBToInt(float r, float g, float b)
{
	r = kClamp(r, 0.0f, 255.0f);
	g = kClamp(g, 0.0f, 255.0f);
	b = kClamp(b, 0.0f, 255.0f);
	return (int(b) << 16) | (int(g) << 8) | int(r);
}

//-------------------------------------------------------------------------------------------

__device__ int kValueToRGB(float v, float a, float b)
{
	if(v >= a)
	{
		if(v <= b)
		{
			v = (v - a) / (b - a);
		}
		else
		{
			v = 1.0f;
		}
	}
	else
	{
		v = 0.0f;
	}
	v *= 255.0f;
	
	int x = int(v);
	return 0xff000000 | (x << 16) | (x << 8) | x;
}

//-------------------------------------------------------------------------------------------

__global__ void kHeightMapToTexture(float *h,unsigned int *img)
{
	int idx = threadIdx.x + __mul24(blockIdx.x, blockDim.x);
	img[idx] = kValueToRGB(h[idx], c_minHeight, c_maxHeight);
}

//-------------------------------------------------------------------------------------------

bool waterLBRenderHeightTexture(WaterLBComputeData *pData)
{
	unsigned int *pOut = (unsigned int *)pData->imageData;
	cudaError err = cudaSuccess;
	
	if(pData->pboId != 0)
	{
		err = cudaGraphicsMapResources(1, &pData->glResource, 0);
		if(checkCUDAResult(err))
		{
			size_t nBytes;
			err = cudaGraphicsResourceGetMappedPointer((void **)&pOut, &nBytes, pData->glResource);
		}
	}
		
	if(checkCUDAResult(err))
	{
		int s = pData->nx * pData->ny;
		dim3 block(576);
		dim3 grid(s / 576);
	
		kHeightMapToTexture<<<grid,block>>>(pData->heightMap, pOut);
		
		err = cudaGetLastError();
		if(checkCUDAResult(err))
		{
			if(pData->textureId != 0)
			{
				err = cudaGraphicsMapResources(1, &pData->glResource, 0);
				if(checkCUDAResult(err))
				{
					cudaArray *pTexture;
					
					err = cudaGraphicsSubResourceGetMappedArray(&pTexture, pData->glResource, 0, 0);
					if(checkCUDAResult(err))
					{
						int tSize = pData->nx * pData->ny * 4;
						
						err = cudaMemcpyToArray(pTexture, 0, 0, pData->imageData, tSize, cudaMemcpyDeviceToDevice);
					}
					checkCUDAResult(cudaGraphicsUnmapResources(1, &pData->glResource, 0));
				}
			}
		}
	}

	if(pData->pboId != 0)
	{
		checkCUDAResult(cudaGraphicsUnmapResources(1, &pData->glResource, 0));
	}
	return checkCUDAResult(err);
}

//-------------------------------------------------------------------------------------------

bool waterLBSetupTextureOutput(WaterLBComputeData *pData,unsigned int texId)
{
	bool res = false;
	int size = pData->nx * pData->ny;
	pData->textureId = texId;
	if(checkCUDAResult(cudaMalloc((void**)&pData->imageData, size * 4)))
	{
		if(checkCUDAResult(cudaGraphicsGLRegisterImage(&pData->glResource, pData->textureId, GL_TEXTURE_2D, cudaGraphicsMapFlagsWriteDiscard)))
		{
			res = true;
		}
	}
	return res;
}

//-------------------------------------------------------------------------------------------

bool waterLBSetupPBOOutput(WaterLBComputeData *pData, unsigned int pboId)
{
	bool res = false;
	cudaError err;
	
	pData->pboId = pboId;
	err = cudaGraphicsGLRegisterBuffer(&pData->glResource, pData->pboId, cudaGraphicsMapFlagsNone);
	if(checkCUDAResult(err))
	{
		res = true;
	}
	return res;
}

//-------------------------------------------------------------------------------------------
